/**
 * Видеочат
 * @file JavaScript-код видеочата
 * @copyright Любезный В. В. <victor@lubezniy.ru> 2020
 * @license https://www.lubezniy.ru/soft/videochat
 * @version 2.0
 */

"use strict";


/**
 * Флаг отладочного режима. При его установке в true в консоль пойдёт много информации по работе скрипта.
 */
const debug = false;

/**
 * Переменные канала сигнализации
 */

let signalingError;
let signalingInterval;
let role = ''; // роль в мультимедиа-звонке (вызывающий caller или отвечающий callee);


/**
 * Служебные функции канала сигнализации
 */

/**
 * Метод отправки данных POST-запросом на сервер.
 *
 * Аргументы:
 *
 *  @param {object} data - объект с заполненными свойствами type, target, msgType, msgData в соответствии с описанием API
 *
 * @return {bool|object} Возвращается false при ошибке (заполняется свойство error) или JSON-объект с ответом от сервера.
 *
 */
function signalingSendData(data) {
    signalingError = null;
    var req = new XMLHttpRequest();
    var body = 'type=' + encodeURIComponent(data.type) + '&target=' + encodeURIComponent(data.target) + '&msgType=' + encodeURIComponent(data.msgType) + '&msgData=' + encodeURIComponent(data.msgData);
    var result = null;
    req.open('POST', './chat', false);
    req.setRequestHeader('Content-Type', 'application/x-www-form-urlencoded');
    req.onreadystatechange = function() {
        if (req.readyState == 4) {
            if (req.status == 200) {
                result = JSON.parse(req.responseText);
            }   else {
                signalingError = req.responseText;
                result = false;
            }
        }
    }
    req.send(body);
    if (signalingError) {
        if (debug) {
            console.log(signalingError);
        }
    }
    return result;
}

/**
 * Метод передачи данных в асинхронном режиме (метод завершится, как только сервер примет сообщение от нас).
 *
 * Аргументы:
 *
 *  @param {string} target - уникальный идентификатор пользователя, которому передаются данные через сервер;
 *  @param {number} msgType - тип сообщения ((0 - текст, 1 - звонок, 2 - приём звонка, 3 - отклонение звонка,
 *      4 - offer на установку прямой медиасвязи, 5 - ответ на offer, 6 - ICE-кандидат,
 *      7 - сообщение о завершении звонка, 8 - отказ пользователя от исходящего звонка до его начала))
 *  @param {string} aData - передаваемое сообщение.
 *
 * Возвращаемое значение отсутствует.
 *
 */
function signalingAsyncSend(target, msgType, aData) {
    var result = signalingSendData({
        type: 'asynctransmit',
        target: target,
        msgType: msgType,
        msgData: aData
    });
    if (result === false) {
        if (debug) {
            console.log('Data sending error: ' + signalingError);
        }
    } else if (result.code != 0) {
        signalingError = result.message;
        if (debug) {
            console.log('Data sending error: ' + signalingError);
        }
    } else {
        signalingDataSent(target, msgType, result.timestamp, aData);
    }
}

/**
 * Метод передачи данных в асинхронном режиме (метод завершится только тогда, когда сервер отдаст сообщение собеседнику).
 *
 * Аргументы:
 *
 *  @param {string} target - уникальный идентификатор пользователя, которому передаются данные через сервер;
 *  @param {number} msgType - тип сообщения ((0 - текст, 1 - звонок, 2 - приём звонка, 3 - отклонение звонка,
 *      4 - offer на установку прямой медиасвязи, 5 - ответ на offer, 6 - ICE-кандидат,
 *      7 - сообщение о завершении звонка, 8 - отказ пользователя от исходящего звонка до его начала))
 *  @param {string} aData - передаваемое сообщение.
 *
 * Возвращаемое значение отсутствует.
 *
 */
function signalingSyncSend(target, msgType, aData) {
    var result = signalingSendData({
        type: 'synctransmit',
        target: target,
        msgType: msgType,
        msgData: aData
    });
    if (result === false) {
        if (debug) {
            console.log('Data sending error: ' + signalingError);
        }
    } else if (result.code != 0) {
        signalingError = result.message;
        if (debug) {
            console.log('Data sending error: ' + signalingError);
        }
    } else {
        signalingDataSent(target, msgType, result.timestamp, aData);
    }
}

/**
 * Метод выполнения проверки входящих сообщений.
 *
 * Аргументы и возвращаемое значение отсутствуют. При получении данных вызывается метод signalingDataReceived.
 *
 */
function signalingCheckData() {
    signalingError = null;
    var req = new XMLHttpRequest();
    var result = null;
    req.open('GET', './newmessages', false);
    req.onreadystatechange = function() {
        if (req.readyState == 4) {
            if (req.status == 200) {
                result = JSON.parse(req.responseText);
            } else {
                signalingError = 'HTTP error ' + req.status;
                result = false;
            }
        }
    }
    req.send();
    if ((result === false) || (result === null)) {
        if (debug) {
            console.log('Data checking error: ' + signalingError);
        }
    } else {
        if (result.code != 0) {
            signalingError = result.message;
            if (debug) {
                console.log('Data checking error: ' + signalingError);
            }
        } else if (result.data.length > 0) {
            for (var i = 0; i < result.data.length; i++) {
                signalingDataReceived(result.data[i].sender, result.data[i].recipient, result.data[i].when, result.data[i].msgType, result.data[i].text);
            }
        }
    }
}



/**
 * Функции для различных задач, связанных с WebRTC
 */

/**
 * Функция отправки offer-а собеседнику.
 * Аргументы:
 *
 *  @param {string} recipientUID - уникальный идентификатор пользователя, которому нужно передать offer.
 *  @param {string} data - SDP-данные offer-а.
 *
 * Возвращаемое значение отсутствует.
 */
function sendOffer(recipientUID, data)
{
    if (debug) {
        console.log('sendOffer is called');
    }
    // Отправку нужно осуществлять в синхронном режиме, чтобы удержать наш RTCPeerConnection от преждевременной
    // отправки ICE-кандидатов (пока удалённый абонент не обработает offer, он не сможет их принять).
    signalingSyncSend(recipientUID, 4, data);
}


/**
 * Функция для создания и отправки собеседнику предложения установить мультимедиа-связь (offer).
 * Вызывается собеседником, инициировавшим звонок, после того, как от собеседника
 * получено согласие на приём звонка.
 *
 * Аргументы:
 *
 *  @param {string} recipientUID - уникальный идентификатор пользователя, которому нужно передать offer.
 *
 * Возвращаемое значение отсутствует.
 */
function createSendOffer(recipientUID)
{
    if (debug) {
        console.log('createSendOffer is called');
    }
    // Создаём offer, используя данные из глобальной переменной offerOptions
    peerConnection.createOffer(offerOptions)
        .then(function(offer) {
            // Offer успешно создан. Устанавливаем его как local description на наш объект класса RTCPeerConnection
            peerConnection.setLocalDescription(offer)
                .then(function() {
                    // Успешно установили local description. Теперь нужно передать информацию из этого offer нашему собеседнику,
                    // чтобы тот проставил её у себя как remote description, сформировал для нас ответ и передал его нам.
                    sendOffer(recipientUID, offer.sdp);
                }).catch(function(reason) {
                   // setLocalDescription не удался - пишем сообщение в лог консоли нашего браузера.
                   if (debug) {
                       console.log('setLocalDescription failed: ' + reason);
                   }
                });
        }).catch(function(reason) {
            // Не удалось создать offer - тоже пишем соответствующее сообщение в лог консоли.
            if (debug) {
                console.log('createOffer failed: ' + reason);
            }
        });
}

/**
 * Функция для обработки принятого offer-а.
 * Аргументы:
 *
 *  @param {string} senderUID - уникальный идентификатор пользователя, от которого получен offer;
 *  @param {string} data - sdp-сообщение с offer-ом.
 *
 * Возвращаемое значение отсутствует.
 */
function receiveOffer(senderUID, data)
{
    if (debug) {
        console.log('receiveOffer is called');
    }
    // Создаём описание сессии на основании полученных данных.
    var desc = new RTCSessionDescription({
        type: 'offer',
        sdp: data
    });

    // устанавливаем на наш peerConnection описание сессии как remoteDescription (сессия от собеседника).
    peerConnection.setRemoteDescription(desc)
        .then(function() {
            // Установили. Теперь добавляем в peerConnection наши локальные треки, которые мы получили ещё когда согласились принять звонок.
            peerConnection.getSenders().forEach(track => peerConnection.removeTrack(track));
            localStream.getTracks().forEach(track => peerConnection.addTrack(track, localStream));
            // Cоздаём ответ на принятый оффер и передаём его дальше, в последующий код.
            return peerConnection.createAnswer();
        }).then(function(answer) {
            // Устанавливаем local description нашего peerConnection из сформированного ответа и также передаём его в последующий код.
            return peerConnection.setLocalDescription(answer);
        }).then(function() {
            // Также отправляем сформированный нами ответ - уже в асинхронном режиме.
            // Это нужно, чтобы наш собеседник прописал наши параметры мультимедиа у себя и мог получать от нас аудиовидеопотоки.
            signalingAsyncSend(senderUID, 5, peerConnection.localDescription.sdp);
        });
}

/**
 * Функция для обработки принятого ответа на наш offer.
 * Аргументы:
 *
 *  @param {string} data - sdp-сообщение с ответом.
 *
 * Возвращаемое значение отсутствует.
 */
function receiveAnswer(data)
{
    if (debug) {
        console.log('receiveAnswer is called');
    }
    // Формируем объект описания сессии
    var sd = new RTCSessionDescription({
        type: 'answer',
        sdp: data
    });
    // Устанавливаем описание сессии в виде remoteDescription на наш peerConnection
    peerConnection.setRemoteDescription(sd);
}

/**
 * Метод обработки согласия вызываемого абонента принять наш звонок
 * @param {string} sender - uid вызываемого абонента
 *
 */
function calleeIsAgree(sender)
{
    // Устанавливаем свою роль как вызывающий
    role = 'caller';
    // Скрываем окно исходящего вызова и показываем окно выполнения видеозвонка.
    $('#callWaiting').dialog('close');
    // Перед показом окна видеозвонка заполняем в диалоге соотв. поле именем собеседника
    var i = contactIndexInList(sender);
    if (i !== false) {
        $('#callWindow').dialog('option', 'title', contacts.getElementsByTagName('li')[i].getAttribute('displayName'));
    } else {
        $('#callWindow').dialog('option', 'title', 'Собеседник');
    }
    $('#callWindow').dialog('open');
    // getUserMedia у нас уже вызван и peerConnection уже создан
    // Добавляем треки из локального потока в peerConnection для последующей передачи собеседнику
    peerConnection.getSenders().forEach(track => peerConnection.removeTrack(track));
    localStream.getTracks().forEach(track => peerConnection.addTrack(track));
    // Если это не будет сделано в событии onNegotiationNeeded, создаём и отправляем offer на установку мультимедиа-связи.
    if (peerConnection.signalingState != 'stable') {
        createSendOffer(sender);
    }
}


/**
 * Функция создания объекта мультимедиа-подключения типа RTCPeerConnection и других объектов мультимедиа-связи (локальный и удалённый медиапотоки).
 * Аргументы отсутствуют.
 * Возвращаемое значение отсутствует.
 */
function createRTCPeerConnection()
{
    if (debug) {
        console.log('createRTCPeerConnection is called');
    }
    // Создаём RTCPeerConnection в глобальной переменной peerConnection
    peerConnection = new RTCPeerConnection(mediaConfig);
    // Подключаем к созданному объекту функции событий
    // negotiationneeded (нужна переотправка собеседнику offer-а на установку мультимедиа-связи)
    peerConnection.addEventListener('negotiationneeded', onNegotiationNeeded);
    // icecandidate (нужно отправить собеседнику информацию о способе связи с нами)
    peerConnection.addEventListener('icecandidate', onICECandidate);
    // iceconnectionstatechange (справочное событие)
    peerConnection.addEventListener('iceconnectionstatechange', onICEConnectionStateChange);
    // track - получен трек для воспроизведения
    peerConnection.addEventListener('track', onTrack);
    // Создаём медиапоток для локального мультимедиа (объект типа MediaStream), если таковой уже не создан.
    // Результат создания сохраняем в глобальной переменной localStream.
    if (!localStream) {
        localStream = new MediaStream;
    }
    // Создаём такой же объект для удалённого потока в глобальной переменной remoteStream
    if (!remoteStream) {
        remoteStream = new MediaStream;
        // Созданный объект для удалённого медиапотока сразу подключаем к элементу страницы video, который будет этот поток воспроизводить.
        // Этот элемент у нас записан в глобальную переменную remoteVideo.
        remoteVideo.srcObject = remoteStream;
    }
}

/**
 * События объекта RTCPeerConnection
 */

/**
 * Событие icecandidate.
 * Вызывается, когда RTCPeerconnection хочет предложить собеседнику способ установки прямой связи через сеть для передачи медиатрафика.
 * Аргументы:
 *
 *  @param {object} event - объект события.
 *
 * Возвращаемое значение отсутствует.
 */
function onICECandidate(event)
{
    if (debug) {
        console.log('RTCPeerConnection.icecandidate is raised');
    }
    if (event.candidate) {
        // ICE-кандидат не пуст - передаём его нашему собеседнику в асинхронном режиме
        signalingAsyncSend(remoteContact.getAttribute('userUID'), 6, JSON.stringify(event.candidate));
    }
}

/**
 * Событие track.
 * Говорит о том, что от собеседника получен мультимедиа-трек, который мы можем воспроизводить у себя (для этого мы данное событие и обрабатываем).
 * Аргументы:
 *
 *  @param {object} event - объект события.
 *
 * Возвращаемое значение отсутствует.
 */
function onTrack(event)
{
    if (debug) {
        console.log('RTCPeerConnection.track is raised');
    }
    // Подключаем полученный трек к нашему медиапотоку удалённой стороны для воспроизведения.
    remoteStream.addTrack(event.track);
}

/**
 * Событие iceconnectionstatechange
 * Для нас справочное. В принципе для работы оно не нужно, но при ошибке нам всё же стоит показать в консоли браузера, что соединение отвалилось.
 * Аргументы:
 *
 *  @param {object} event - объект события.
 *
 * Возвращаемое значение отсутствует.
 */
function onICEConnectionStateChange(event)
{
    if (debug) {
        console.log('RTCPeerConnection.icestatechange is raised');
    }
    if (peerConnection.iceConnectionState === 'failed') {
        // Соединение в статусе failed - выдаём в консоль.
        console.log('ICE connection failed');
    }
}

/**
 * Событие negotiationneeded.
 * Вызывается в случае, когда надо отправить или переотправить offer (если есть, через что).
 * Аргументы:
 *
 *  @param {object} event - объект события.
 *
 * Возвращаемое значение отсутствует.
 */
function onNegotiationNeeded()
{
    if (debug) {
        console.log('RTCPeerConnection.negotiationneeded is raised');
    }
    if (peerConnection.signalingState != 'stable') {
        // Проблема с каналом сигнализации по протоколу ICE - делать ничего не нужно.
        if (debug) {
            console.log('RTCPeerConnection.signalingState = ' + peerConnection.signalingState);
        }
        return;
    }
    if (role == 'caller') {
        // Если звоним мы, то отправляем собеседнику оффер (при необходимости пересоздав его).
        if (peerConnection.localDescription) {
            sendOffer(remoteContact.getAttribute('userUID'), peerConnection.localDescription.sdp);
        } else {
            createSendOffer(remoteContact.getAttribute('userUID'));
        }
    }
}


/**
 * Вспомогательные функции для выполнения различных действий
 */

/**
 * Очистка объекта от дочерних узлов
 * @param {object} object
 */
function clearChildNodes(object)
{
    while(object.childNodes.length > 0) {
        object.removeChild(object.childNodes[0]);
    }
}

/**
 * Получение и заполнение списка контактов.
 * Аргументы отсутствуют.
 * Возвращаемое значение отсутствует.
 */
function getContactsList()
{
    // формируем и осуществляем GET-запрос на сервер
    var req = new XMLHttpRequest();
    var result = null;
    req.open ('GET', '/contacts', false);
    req.onreadystatechange = function() {
        if (req.readyState == 4) {
            if (req.status == 200) {
                result = JSON.parse(req.responseText);
            }   else {
                result = null;
            }
        }
    }
    req.send();
    // заполняем текущие данные
    if (result) {
        // есть ответ с http-кодом 200
        if (result.code == 1) {
            // пользователь не авторизован - редиректим на главную, дальше само пойдёт на логин
            document.location.href = '/';
        } else if (result.code == 0) {
            // заполняем данные текущего пользователя
            clearChildNodes(currentUser);
            // устанавливаем атрибуты на элемент текущего залогиненного пользователя, чтобы потом можно было легко брать значения
            currentUser.setAttribute('UID', result.current.UID);
            currentUser.setAttribute('displayName', result.current.displayName);
            currentUser.appendChild(document.createTextNode(result.current.displayName));
            // заполняем данные списка контактов с учётом уже имеющегося
            if (result.data.length > 0) {
                // перебор в цикле
                for (var i = 0; i < result.data.length; i++) {
                    var index = contactIndexInList(result.data[i].uid);
                    if (index === false) {
                        // пользователя в списке контактов нет - создаём нового
                        var el = document.createElement('li');
                        el.setAttribute('userUID', result.data[i].uid);
                        el.setAttribute('displayName', result.data[i].displayName);
                        el.setAttribute('onClick', 'setChat(this)');
                        el.setAttribute('online', (result.data[i].online ? 'yes' : 'no'));
                        // В зависимости от того, онлайн пользователь или оффлайн, ставим класс на элемент
                        el.setAttribute('class', (result.data[i].online ? 'online' : 'offline'));
                        el.appendChild(document.createTextNode(' ' + result.data[i].displayName));
                        contacts.appendChild(el);
                    } else {
                        // пользователь в списке есть - обновляем данные
                        var el = contacts.getElementsByTagName('li')[index];
                        el.setAttribute('userUID', result.data[i].uid);
                        el.setAttribute('displayName', result.data[i].displayName);
                        el.setAttribute('online', (result.data[i].online ? 'yes' : 'no'));
                        el.setAttribute('class', (result.data[i].online ? 'online' : 'offline'));
                    }
                }
            }
        }
    }
}

/**
 * Метод проверки существования контакта в списке.
 * Аргументы:
 *
 * @param {string} uid - уникальный идентификатор пользователя (контакта).
 * @return {bool|number} Возвращается индекс контакта в списке или false, если контакт не найден.
 */
function contactIndexInList(uid)
{
    // ищем элементы в нашем ul списка контактов
    var nodes = contacts.getElementsByTagName('li');
    if (nodes.length == 0) {
        // список по какой-то причине пуст - сразу возвращаем false
        return false;
    }
   // в цикле перебираем список пока
   for (var i = 0; i < nodes.length; i++) {
       if (nodes[i].getAttribute('userUID') == uid) return i;
   }
   // цикл закончился ничем - возвращаем false
   return false;
}

/**
 * Установка в списке контактов информации о наличии новых сообщений
 * Аргументы:
 *
 *  @param {string} uid - уникальный идентификатор контакта, которому нужно проставить информацию.
 *
 * Возвращаемое значение отсутствует.
 */
function setContactNewMessages(uid)
{
    if (debug) {
        console.log('setContactNewMessages is called');
    }
    // в массив nodes кладём все элементы списка контактов
    var nodes = contacts.getElementsByTagName('li');
    if (nodes.length > 0) {
        // список контактов не пуст, перебираем элементы
        for (var i = 0; i < nodes.length; i++) {
            if (nodes[i].getAttribute('userUID') == uid) {
                // нашли нужный, ставим ему жирный шрифт в качестве отметки в атрибут style (остальные свойства стиля берутся из css)
                nodes[i].setAttribute('style', 'font-weight: bold');
                break;
            }
        }
    }
}

/**
 * Снятие с контакта уведомления о наличии новых сообщений.
 * Аргументы:
 *
 *  @param {string} uid - уникальный идентификатор контакта, которому нужно проставить информацию.
 *
 * Возвращаемое значение отсутствует.
 */
function clearContactNewMessages(uid)
{
    if (debug) {
        console.log('clearContactNewMessages is called');
    }
    // в массив nodes кладём все элементы списка контактов
    var nodes = contacts.getElementsByTagName('li');
    if (nodes.length > 0) {
        // список контактов не пуст, перебираем элементы
        for (var i = 0; i < nodes.length; i++) {
            if (nodes[i].getAttribute('userUID') == uid) {
                // контакт нашли - просто удаляем соотв. элементу атрибут style
                nodes[i].removeAttribute('style');
                break;
            }
        }
    }
}

/**
 * Добавление записи в отображаемую историю сообщений.
 *
 * Аргументы:
 *
 *  @param {string} author - уникальный идентификатор пользователя - автора сообщения (либо текущий залогиненный пользователь, либо собеседник);
 *  @param {number} time - UNIX-timestamp даты и времени приёма сообщения сервером (кол-во секунд, прошедшее с 00:00 1 января 1970 года до заданного времени);
 *  @param {string} text - текст сообщения.
 *
 * Возвращаемое значение отсутствует.
 */
function addMessageToHistory(author, time, text)
{
    if (debug) {
        console.log('addMessageToHistory is called');
    }
    // Создаём html-элемент div - общий элемент для всей записи истории.
    var el = document.createElement('div');
    // Устанавливаем созданному элементу класс historyMessage для применения соответствующего стиля из css.
    el.setAttribute('class', 'historyMessage');
    // Создаём html-элемент span для отображения автора сообщения.
    var a = document.createElement('span');
    // Устанавливаем созданному элементу span класс author для применения соответствующего стиля из css.
    a.setAttribute('class', 'author');
    // Получаем отображаемое имя автора сообщения по уникальному идентификатору и записываем его в переменную s.
    if (author == currentUser.getAttribute('uid')) {
        // Автор - текущий пользователь. Берём имя из соответствующего атрибута элемента currentUser.
        var s = currentUser.getAttribute('displayName');
    } else {
        // автор - наш собеседник. Берём имя из соответствуюшего атрибута элемента remoteContact.
        var s = remoteContact.getAttribute('displayName');
    }
    // Добавляем к отображаемому имени дату и время сообщения, преобразовав их в человекочитаемый формат.
    s += ', ' + getDatetimeFromTimestamp(time);
    // В конец добавляем двоеточие с пробелом.
    s += ': ';
    // Добавляем получившийся текст в наш "авторский" span.
    a.appendChild(document.createTextNode(s));
    // Получившийся span, в свою очередь, добавляем в div.
    el.appendChild(a);
    // Создаём ещё один элемент span - уже для текста сообщения.
    a = document.createElement('span');
    // Устанавливаем этому элементу класс message для применения стиля.
    a.setAttribute('class', 'message');
    // Добавляем к этому элементу собственно текст сообщения.
    a.appendChild(document.createTextNode(text));
    // Добавляем созданный span в div после автора.
    el.appendChild(a);
    // Добавляем получившийся div со всем содержимым в блок истории сообщений.
    msgHistory.appendChild(el);
}

/**
 * Получение и отображение истории сообщений по заданному контакту.
 * Аргументы:
 *
 *  @param {string} contactUID - уникальный идентификатор пользователя - нашего собеседника
 *
 * Возвращаемое значение отсутствует.
 */
function getHistory(contactUID)
{
    if (debug) {
        console.log('getHistory is called');
    }
    // Очищаем блок, отображающий историю сообщений, от дочерних элементов.
    clearChildNodes(msgHistory);
    // Делаем синхронный HTTP-запрос к серверу на предмет получения истории сообщений с помощью обычного XmlHttpRequest.
    var req = new XMLHttpRequest();
    var result = null;
    req.open('GET', '/history?contact=' + encodeURIComponent(contactUID), false);
    req.onreadystatechange = function() {
        if (req.readyState == 4) {
            if (req.status == 200) {
                result = JSON.parse(req.responseText);
            }   else {
                result = null;
            }
        }
    }
    req.send();
    // Запрос окончен; смотрим, что у нас там получилось
    if (result) {
        if (result.code == 1) {
            // Сервер вернул ошибку - пользователь не авторизован. Перенаправляем пользователя на страницу для решения этого вопроса.
            document.location.href = '/';
        } else if (result.code == 0) {
            // Сервер вернул успешный результат - обрабатываем.
            if (result.data.length > 0) {
                // В истории сообщений есть элементы - перебираем их и добавляем.
                for (var i = 0; i < result.data.length; i++) {
                    if (result.data[i].msgType == 0) {
                        // Сообщение чата - добавляем в историю. Остальные игнорируем.
                        addMessageToHistory(result.data[i].sender, result.data[i].when, result.data[i].text);
                    }
                }
            }
        }
    }
}

/**
 * Преобразование даты и времени из timestamp в текст.
 * Аргументы:
 *
 *  @param {number} stamp - дата-время в виде UNIX timestamp (кол-во секунд с 00:00 1 января 1970 года)
 *
 * @return {string} Возвращается дата и время в виде строки - на текущий момент так, как выдаёт его браузер.
 */
function getDatetimeFromTimestamp(stamp)
{
    var d = new Date(stamp * 1000);
    return d.toString();
}

/**
 * События на канале сигнализации (через который передаются текстовый чат и информация для установки и завершения мультимедиа-связи).
 */

/**
 * Событие окончания отправки данных.
 * Аргументы:
 *
 *  @param {string} target - уникальный идентификатор пользователя - нашего собеседника (которому были отправлены данные);
 *  @param {number} msgType - тип переданного сообщения (0 - текст, 1 - звонок, 2 - приём видеозвонка, 3 - отклонение звонка,
 *      4 - offer, 5 - ответ на offer, 6 - ICE-кандидат, 7 - сообщение о завершении звонка,
 *      8 - отказ пользователя от исходящего звонка до его начала, 9 - запрос на автоперезвон);
 *  @param {number} time - UNIX-timestamp даты и времени приёма сообщения сервером (кол-во секунд, прошедшее с 00:00 1 января 1970 года до заданного времени);
 *  @param {string} text - переданные данные.
 *
 * Возвращаемое значение отсутствует.
 */
function signalingDataSent(target, msgType, time, text)
{
    if (debug) {
        console.log('signalingDataSent is called');
    }
    if (msgType == 0) {
        // Это было отправлено сообщение чата.
        // Добавляем его в текущую историю сообщений.
        addMessageToHistory(currentUser.getAttribute('uid'), time, text);
        // Очищаем значение поля сообщения в форме его отправки.
        sendMsgForm.reset();
    }
}

/**
 * Событие обработки приёма входящего сообщения.
 * Аргументы:
 *  @param {string} sender - уникальный идентификатор пользователя - отправителя сообщения;
 *  @param {string} recipient - уникальный идентификатор пользователя - получателя сообщения (т. е., наш);
 *  @param {number} when - UNIX-timestamp даты и времени приёма сообщения сервером (кол-во секунд, прошедшее с 00:00 1 января 1970 года до заданного времени);
 *  @param {number} msgType - тип переданного сообщения (0 - текст, 1 - звонок, 2 - приём видеозвонка, 3 - отклонение звонка,
 *      4 - offer, 5 - ответ на offer, 6 - ICE-кандидат, 7 - сообщение о завершении звонка,
 *      8 - отказ пользователя от исходящего звонка до его начала, 9 - запрос на автоперезвон);
 *  @param {string} text - принятые данные.
 * Возвращаемое значение отсутствует.
 */
function signalingDataReceived(sender, recipient, when, msgType, text)
{
    if (debug) {
        console.log('signalingDataReceived is called, msgType ' + msgType + ', text ' + text);
    }
    // Все действия выполняем в зависимости от типа принятого сообщения.
    if (msgType == 0) {
        // Пришло сообщение чата.
        // Если история у нас активна для отправителя сообщения, добавляем сообщение в неё.
        if (sender == remoteContact.getAttribute('userUID')) {
            addMessageToHistory(sender, when, text)
        } else {
            // Иначе просто отмечаем в списке контактов, что у такого-то пользователя есть для Вас непрочитанные сообщения.
            setContactNewMessages(sender);
        }
        // Для этого типа сообщения всё сделали, завершаемся
        return;
    }

    if (msgType == 1) {
        // Поступил входящий звонок
        // Первое - по наличию открытых соответствующих диалогов проверяем, не заняты ли мы уже общением или ожиданием вызова.
        if ($('#callDialog').dialog('isOpen') || $('#callWaiting').dialog('isOpen') || $('#callWindow').dialog('isOpen')) {
            // Заняты; отправляем обратно сообщение о том, что отклоняем звонок, и завершаем функцию.
            if (debug) {
                console.log('Incoming call is automatically declined.');
            }
            signalingAsyncSend(sender, 3, 'Абонент занят.');
            return;
        }
        // Теперь нужно получить отображаемое имя вызывающего и отобразить диалог приёма входящего звонка с этим именем.
        // Получаем номер (индекс) отправителя в списке контактов.
        var i = contactIndexInList(sender);
        // На всякий случай убеждаемся, что пользователь есть у нас в списке.
        if (i !== false) {
            // Записываем отображаемое имя в соответствующий элемент диалога приёма звонка.
            callMessage.innerText = contacts.getElementsByTagName('li')[i].getAttribute('displayName');
            callMessage.setAttribute('displayName', contacts.getElementsByTagName('li')[i].getAttribute('displayName'));
            callMessage.setAttribute('userUID', sender);
            // Отображаем диалог.
            $('#callDialog').dialog('open');
            // Запускаем воспроизведение рингтона.
            ring.play();
        }
        return;
    }

    if (msgType == 2) {
        // Наш собеседник согласился принять от нас входящий звонок.
        calleeIsAgree(sender);
        return;
    }

    if (msgType == 3) {
        // Наш собеседник отказался принять наш входящий звонок.
        if (sender == remoteContact.getAttribute('userUID')) {
            // Если наш абонент ещё находится в списке контактов собеседника, добавляем сообщение в историю.
            addMessageToHistory(sender, when, 'Звонок отклонён. ' + text);
        } else {
            // Иначе просто отмечаем в списке контактов, что у такого-то пользователя есть для Вас непрочитанные сообщения.
            setContactNewMessages(sender);
        }
        // Скрываем окно исходящего вызова.
        $('#callWaiting').dialog('close');
        // Убираем лишние объекты
        localVideo.srcObject = null;
        remoteVideo.srcObject = null;
        localStream = null;
        remoteStream = null;
        return;
    }

    if (msgType == 4) {
        // От собеседника поступил offer на установку мультимедиа-связи. Нужно его обработать.
        receiveOffer(sender, text);
        return;
    }

    if (msgType == 5) {
        // От собеседника поступил ответ на полученный от нас offer
        receiveAnswer(text);
        return;
    }

    if (msgType == 6) {
        // Поступила информация об ICE-кандидате (способе установки прямой мультимедиа-связи)
        // Просто добавляем её в peerConnection
        peerConnection.addIceCandidate(JSON.parse(text));
        return;
    }

    if (msgType == 7) {
        // От собеседника поступило сообщение о завершении звонка.
        // Закрываем соединение с собеседником
        peerConnection.close();
        // Отключаем воспроизведение
        localVideo.pause();
        remoteVideo.pause();
        // останавливаем все треки локального потока
        if (localStream !== null) {
            localStream.getTracks().forEach(track => track.stop());
        }
        // Убираем лишние объекты
        localVideo.srcObject = null;
        remoteVideo.srcObject = null;
        localStream = null;
        remoteStream = null;
        // Сбрасываем роль
        role = '';
        // Скрываем окно выполнения звонка.
        $('#callWindow').dialog('close');
        return;
    }

    if (msgType == 8) {
        // Поступил отказ вызывающего пользователя от совершения исходящего звонка (звонящий положил трубку до того, как мы ответили).
        // Останавливаем воспроизведение рингтона
        ring.pause();
        // Скрываем окно входящего вызова.
        $('#callDialog').dialog('close');
        return;
    }

    if (msgType == 9) {
        // Вызываемый просит нас повторить соединение
        // закрываем соединение
        peerConnection.close();
        // заново отправляем offer
        calleeIsAgree(sender);
        return;
    }
    // Если в будущем потребуется обрабатывать ещё какие-то типы сообщений, можно дописать код обработки сюда.
}

/**
 * Событие изменения списка аудиовидеоустройств
 */
function fillMediaDeviceList()
{
    navigator.mediaDevices.enumerateDevices()
        .then(function(devices) {
            // очищаем списки аудиовидеоустройств
            clearChildNodes(currentAudioDevice);
            clearChildNodes(currentVideoDevice);
            // перебираем устройства в цикле
            devices.forEach(function(device) {
                // делим строку параметров устройства
                let [kind, type, direction] = device.kind.match(/(\w+)(input|output)/i);
                // выводим только устройства ввода
                if (direction == 'input') {
                    var element = document.createElement('option');
                    element.appendChild(document.createTextNode(device.label));
                    element.setAttribute('value', device.deviceId);
                    if (type === "audio") {
                        currentAudioDevice.appendChild(element);
                    } else if (type === "video") {
                        currentVideoDevice.appendChild(element);
                    }
                }
            });
        });
}


/**
 * События на объектах страницы
 */

/**
 * Событие onLoad на body - создание объектов и запуск проверки наличия новых сообщений.
 * Аргументы отсутствуют.
 * Возвращаемое значение отсутствует.
 */
function start()
{
    if (debug) {
        console.log('start is called');
    }
    try {
        // Задаём переменные для упрощения доступа к элементам страницы.
        callDialog = document.getElementById('callDialog');
        callMessage = document.getElementById('callMsg');
        callWaiting = document.getElementById('callWaiting');
        callWaitingMessage = document.getElementById('callWaitingMsg');
        chatData = document.getElementById('chatData');
        defaultChatData = document.getElementById('defaultChatData');
        contacts = document.getElementById('contacts');
        currentUser = document.getElementById('currentUser');
        remoteContact = document.getElementById('remoteContact');
        msgHistory = document.getElementById('history');
        sendMsgForm = document.getElementById('sendMsgForm');
        mediaCallForm = document.getElementById('mediaCallForm');
        callWindow = document.getElementById('callWindow');
        incomingCallForm = document.getElementById('incomingCallForm');
        localVideo = document.getElementById('localVideo');
        remoteVideo = document.getElementById('remoteVideo');
        ring = document.getElementById('ring');
        currentAudioDevice = document.getElementById('currentAudioDevice');
        currentVideoDevice = document.getElementById('currentVideoDevice');
        // Получаем и отображаем список пользователей и текущие данные.
        getContactsList();
        // Устанавливаем событие на изменение списка устройств
        navigator.mediaDevices.addEventListener('devicechange', fillMediaDeviceList);
        // Включаем обновление списка контактов каждые 5 секунд.
        getUsersInterval = setInterval(() => { getContactsList(); }, 5000);
        // Создаём объект канала сигнализации, подключаем к нему события datasent (данные отправлены) и datareceived (есть входящие данные) и запускаем ежесекундную проверку новых сообщений.
        signalingInterval = setInterval(() => { signalingCheckData(); }, 1000);
    } catch(e) {
        alert(e);
    }
}

/**
 * Событие onClick на элементе списка контактов.
 * Также вызывается при приёме входящего звонка, чтобы переключиться на текстовый чат с собеседником
 * и обеспечить обмен служебными сообщениями при организации канала мультимедиа-связи.
 *
 * Аргументы:
 *
 *  @param {object} element - элемент типа li, в котором должны быть установлены также атрибуты userUID (уникальный идентификатор контакта), displayName (отображаемое имя контакта) и online (статус пользователя).
 *
 * Возвращаемое значение отсутствует.
 */
function setChat(element)
{
    if (debug) {
        console.log('setChat is called');
    }
    // Задача - отобразить контакт в поле беседы, скачать и отобразить историю сообщений с контактом.
    // Устанавливаем атрибуты нашего собеседника как текущего в элементе, представленном переменной remoteContact
    remoteContact.setAttribute('userUID', element.getAttribute('userUID'));
    remoteContact.setAttribute('displayName', element.getAttribute('displayName'));
    remoteContact.setAttribute('online', element.getAttribute('online'));
    // Делаем отображение элемента из переменной remoteContact.
    // Сначала удаляем дочерние узлы, если таковые есть.
    clearChildNodes(remoteContact);
    // Теперь добавляем дочерний текстовый узел с отображаемым именем.
    remoteContact.appendChild(document.createTextNode(element.getAttribute('displayName')));
    // Получаем историю сообщений с данным контактом.
    getHistory(element.getAttribute('userUID'));
    // Делаем блок истории сообщений видимым.
    msgHistory.style.display = 'block';
    // Убираем уведомления о новых сообщениях по контакту (открыли - значит, прочитали).
    clearContactNewMessages(element.getAttribute('userUID'));
    // Делаем видимым диалог чата с собеседником.
    defaultChatData.style.display = 'none';
    chatData.style.display = 'table-cell';
}

/**
 * Событие щелчка по кнопке отправки текстового сообщения в чат.
 * Аргументы и возвращаемое значение отсутствуют.
 */
function sendChatMessage()
{
    if (debug) {
        console.log('sendChatMessage is called');
    }
    // Отправляем сообщение из поля формы отправки в асинхронном режиме.
    signalingAsyncSend(remoteContact.getAttribute('userUID'), 0, sendMsgForm.message.value);
}

/**
 * Событие щелчка по кнопке "Вызвать" для вызова абонента по аудио-видео
 * Аргументы и возвращаемое значение отсутствуют.
 */
function makeCall()
{
    if (debug) {
        console.log('makeCall is called');
    }
    // Проверяем установку флажков на форме (должен быть установлен хотя бы один из двух - аудио, видео).
    if (!mediaCallForm.audio.checked && !mediaCallForm.video.checked) {
        // Флажки не установлены - ругаемся и завершаем обработку.
        alert('Нужно выбрать аудио и (или) видео.');
        return;
    }
    // Задаём конфигурацию медиаустройств в соответствии с установкой флажков.
    mediaConfig.audio = mediaCallForm.audio.checked;
    mediaConfig.video = mediaCallForm.video.checked;
    // Создаём объекты для мультимедиа-соединения и медиапотоков.
    createRTCPeerConnection();
    // Запрашиваем у браузера доступ к аудиовидеоустройствам в соответствии с конфигурацией и получаем локальный медиапоток.
    navigator.mediaDevices.getUserMedia(mediaConfig)
        .then(function(stream) {
            // Доступ получили.
            // Устанавливаем свой локальный поток на локальное видео
            localStream = stream;
            localVideo.srcObject = localStream;
            // заполняем текущий список устройств
            fillMediaDeviceList();
            // Показываем окно, что мы кого-то пытаемся вызвать.
            callWaitingMessage.innerText = remoteContact.getAttribute('displayName');
            $('#callWaiting').dialog('open');
            // Отправляем собеседнику запрос на приём исходящего звонка в асинхронном режиме через канал сигнализации.
            signalingAsyncSend(remoteContact.getAttribute('userUID'), 1, '');
        }).catch(function(e) {
            // Ошибка в процессе получения доступа к аудиовидеоустройствам
            // Ругаемся.
            alert('Не удалось открыть устройства мультимедиа. ' + e);
            // Скрываем диалог исходящего вызова.
            $('#callWaiting').dialog('close');
        });
}

/**
 * Событие щелчка по кнопке "Остановить" на форме выполнения исходящего звонка (до того, как собеседник ответил)
 * Аргументы и возвращаемое значение отсутствуют.
 */
function callStop()
{
    if (debug) {
        console.log('callStop is called');
    }
    // Скрываем диалог исходящего вызова.
    $('#callWaiting').dialog('close');
    // Отправляем собеседнику служебное сообщение, что пользователь передумал звонить, через канал сигнализации в асинхронном режиме.
    signalingAsyncSend(remoteContact.getAttribute('userUID'), 8, '');
}

/**
 * Событие щелчка по кнопке "Остановить" на форме видеозвонка (когда звонок в ходе выполнения).
 * Аргументы и возвращаемое значение отсутствуют.
 */
function activeCallStop()
{
    if (debug) {
        console.log('activeCallStop is called');
    }
    // Скрываем диалог вызова.
    $('#callWindow').dialog('close');
    // Отправляем собеседнику служебное сообщение, что мы решили завершить звонок, через канал сигнализации в асинхронном режиме.
    signalingAsyncSend(remoteContact.getAttribute('userUID'), 7, '');
    // Закрываем соединение
    peerConnection.close();
    // останавливаем все треки локального потока
    localStream.getTracks().forEach(track => track.stop());
    // Убираем лишние объекты
    localVideo.srcObject = null;
    remoteVideo.srcObject = null;
    localStream = null;
    remoteStream = null;
    // Сбрасываем роль
    role = '';
}

/**
 * Событие щелчка по кнопке "Принять звонок" на форме приёма входящего звонка.
 * Аргументы и возвращаемое значение отсутствуют.
 */
function callAccept()
{
    if (debug) {
        console.log('callAccept is called');
    }
    // Останавливаем воспроизведение рингтона
    ring.pause();
    // Проверяем установку флажков в окне приёма - какую мультимедиа-информацию мы будем передавать собеседнику.
    if (!incomingCallForm.audio.checked && !incomingCallForm.video.checked) {
        // Ни один флажок не установлен - ругаемся и ждём следующего щелчка, уже с установленным флажком
        alert('Нужно выбрать аудио и (или) видео.');
        return;
    }
    // Задаём конфигурацию медиаустройств в соответствии с установкой флажков.
    mediaConfig.audio = incomingCallForm.audio.checked;
    mediaConfig.video = incomingCallForm.video.checked;
    // Переключаем чат на нашего собеседника, чтобы сделать его текущим.
    var remoteUID = callMessage.getAttribute('userUID');
    setChat(contacts.getElementsByTagName('li')[contactIndexInList(remoteUID)]);
    // Устанавливаем нашу роль как вызываемого
    role = 'callee';
    // Создаём RTCPeerConnection с объектами локального и удалённого медиапотоков
    createRTCPeerConnection();
    // Запрашиваем у браузера доступ к аудиовидеоустройствам в соответствии с конфигурацией и получаем локальный медиапоток.
    navigator.mediaDevices
        .getUserMedia(mediaConfig)
        .then(function(stream) {
            // Устанавливаем полученный локальный поток в переменную и на локальное видео
            localStream = stream;
            localVideo.srcObject = localStream;
            peerConnection.getSenders().forEach(track => peerConnection.removeTrack(track));
            localStream.getTracks().forEach(track => peerConnection.addTrack(track, localStream));
            // заполняем текущий список устройств
            fillMediaDeviceList();
            // Скрываем диалог, на котором мы нажали кнопку.
            $('#callDialog').dialog('close');
            // Перед показом окна видеозвонка заполняем в диалоге соотв. поле именем собеседника
            $('#callWindow').dialog('option', 'title', callMessage.getAttribute('displayName'));
            // Отображаем окно видеозвонка.
            $('#callWindow').dialog('open');
            // Отправляем собеседнику сообщение о согласии принять вызов через канал сигнализации в асинхронном режиме.
            signalingAsyncSend(remoteUID, 2, '');
        }).catch(function(e) {
            // Не удалось получить доступ к аудиовидеоустройствам
            alert('Не удалось открыть устройства передачи. ' + e);
        });
}

/**
 * Событие щелчка по кнопке "Отклонить" на форме приёма входящего звонка.
 * @param {string} message Сообщение, с которым отклоняется звонок.
 * Аргументы и возвращаемое значение отсутствуют.
 */
function callDecline(message)
{
    if (debug) {
        console.log('callDecline is called');
    }
    // Останавливаем воспроизведение рингтона
    ring.pause();
    // Скрываем диалог входящего вызова.
    $('#callDialog').dialog('close');
    // Отправляем собеседнику сообщение об отказе принять вызов через канал сигнализации в асинхронном режиме.
    signalingAsyncSend(callMessage.getAttribute('userUID'), 3, message);
}

/**
 * replace media track
 */
function mediaTrackReplace(track, senders)
{
    if (debug) {
        console.log('Replace track kind ' + track.kind);
    }
    var sender = senders.find(function(s) { return (s.track.kind == track.kind);  });
    sender.replaceTrack(track);
}

/**
 * Метод смены источника медиасигнала
 * @param {string} video id видеоустройства или пустая строка или null
 * @param {string} audio id аудиоустройства или пустая строка или null
 */
function changeMediaDevice(video, audio)
{
    if (debug) {
        console.log('Changing media device: audio ' + audio + ', video ' + video);
    }
    // Задаём конфигурацию медиаустройств в соответствии с установками
    if (audio !== null) {
        if (audio == '') {
            mediaConfig.audio = false;
        } else {
            mediaConfig.audio = {
                deviceId: {exact:  audio}
            };
        }
    }
    if (video !== null) {
        if (video == '') {
            mediaConfig.video = false;
        } else {
            mediaConfig.video = {
                deviceId: {exact:  video}
            };
        }
    }
    // Тормозим текущие треки (обязательно, иначе будут проблемы с заменой)
    localStream.getTracks().forEach(track => track.stop());
    // Запрашиваем у браузера доступ к аудиовидеоустройствам в соответствии с конфигурацией и получаем локальный медиапоток.
    navigator.mediaDevices.getUserMedia(mediaConfig)
        .then(function(stream) {
            // Доступ получили.
            var senders = peerConnection.getSenders();
            // отрабатываем видеотреки
            if (video !== null) {
                stream.getVideoTracks().forEach(track => mediaTrackReplace(track, senders));
            }
            // отрабатываем аудиотреки
            if (audio !== null) {
                stream.getAudioTracks().forEach(track => mediaTrackReplace(track, senders));
            }
            // Устанавливаем свой локальный поток на локальное видео
            localStream = stream;
            localVideo.srcObject = localStream;
        }
    );
}